# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import Schema, fields, post_dump, post_load

from polyaxon_schemas.base import BaseConfig, BaseMultiSchema
from polyaxon_schemas.utils import Tensor


class BaseLossSchema(Schema):
    input_layer = Tensor(allow_none=True)
    output_layer = Tensor(allow_none=True)
    weights = fields.Float(default=1.0, missing=1.0)
    name = fields.Str(allow_none=True)
    collect = fields.Bool(default=True, missing=True)


class BaseLossConfig(BaseConfig):
    REDUCED_ATTRIBUTES = ['input_layer', 'output_layer', 'name']

    def __init__(self, input_layer=None, output_layer=None, weights=1.0, name=None, collect=True):
        self.input_layer = input_layer
        self.output_layer = output_layer
        self.weights = weights
        self.name = name
        self.collect = collect


class AbsoluteDifferenceSchema(BaseLossSchema):
    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return AbsoluteDifferenceConfig(**data)

    @post_dump
    def unmake(self, data):
        return AbsoluteDifferenceConfig.remove_reduced_attrs(data)


class AbsoluteDifferenceConfig(BaseLossConfig):
    """Adds an Absolute Difference loss to the training procedure.

    `weights` acts as a coefficient for the loss. If a scalar is provided, then
    the loss is simply scaled by the given value. If `weights` is a `Tensor` of
    shape `[batch_size]`, then the total loss for each sample of the batch is
    rescaled by the corresponding element in the `weights` vector. If the shape of
    `weights` matches the shape of `predictions`, then the loss of each
    measurable element of `predictions` is scaled by the corresponding value of
    `weights`.

    Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.

    Returns:
        A scalar `Tensor` representing the loss value.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: AbsoluteDifference
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      loss:
        AbsoluteDifference:
          input_layer: labels
          output_layer: dense_out
      # other model properties
    ```
    """
    IDENTIFIER = 'AbsoluteDifference'
    SCHEMA = AbsoluteDifferenceSchema


class MeanSquaredErrorSchema(BaseLossSchema):
    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return MeanSquaredErrorConfig(**data)

    @post_dump
    def unmake(self, data):
        return MeanSquaredErrorConfig.remove_reduced_attrs(data)


class MeanSquaredErrorConfig(BaseLossConfig):
    """Computes Mean Square Loss.

    Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.

    Returns:
        A scalar `Tensor` representing the loss value.

    Raises:
        ValueError: If `predictions` shape doesn't match `labels` shape, or `weights` is `None`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: MeanSquaredError
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      loss:
        MeanSquaredError:
          input_layer: labels
          output_layer: dense_out
      # other model properties
    ```
    """
    IDENTIFIER = 'MeanSquaredError'
    SCHEMA = MeanSquaredErrorSchema


class LogLossSchema(BaseLossSchema):
    epsilon = fields.Float(default=1e-7, missing=1e-7)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return LogLossConfig(**data)

    @post_dump
    def unmake(self, data):
        return LogLossConfig.remove_reduced_attrs(data)


class LogLossConfig(BaseLossConfig):
    """Computes Huber Loss for DQN.

    Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.
        epsilon: A small value.

    Returns:
        A scalar `Tensor` representing the loss value.

    Raises:
        ValueError: If `predictions` shape doesn't match `labels` shape, or `weights` is `None`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: LogLoss
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      loss:
        LogLoss:
          input_layer: labels
          output_layer: dense_out
      # other model properties
    ```
    """
    IDENTIFIER = 'LogLoss'
    SCHEMA = LogLossSchema

    def __init__(self,
                 input_layer=None,
                 output_layer=None,
                 weights=1.0,
                 epsilon=1e-7,
                 name=None,
                 collect=True):
        super(LogLossConfig, self).__init__(input_layer, output_layer, weights, name, collect)
        self.epsilon = epsilon


class HuberLossSchema(BaseLossSchema):
    clip = fields.Float(default=0., missing=0.)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return HuberLossConfig(**data)

    @post_dump
    def unmake(self, data):
        return HuberLossConfig.remove_reduced_attrs(data)


class HuberLossConfig(BaseLossConfig):
    """Computes Huber Loss for DQN.

    [Wikipedia link](https://en.wikipedia.org/wiki/Huber_loss)
    [DeepMind link](https://sites.google.com/a/deepmind.com/dqn/)

    Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.

    Returns:
        A scalar `Tensor` representing the loss value.

    Raises:
        ValueError: If `predictions` shape doesn't match `labels` shape, or `weights` is `None`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: HuberLoss
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      loss:
        HuberLoss:
          clip: 0.2
      # other model properties
    ```
    """
    IDENTIFIER = 'HuberLoss'
    SCHEMA = HuberLossSchema

    def __init__(self,
                 input_layer=None,
                 output_layer=None,
                 weights=1.0,
                 clip=0.,
                 name=None,
                 collect=True):
        super(HuberLossConfig, self).__init__(input_layer, output_layer, weights, name, collect)
        self.clip = clip


class ClippedDeltaLossSchema(BaseLossSchema):
    clip_value_min = fields.Float(default=-1., missing=-1.)
    clip_value_max = fields.Float(default=-1., missing=-1.)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return ClippedDeltaLossConfig(**data)

    @post_dump
    def unmake(self, data):
        return ClippedDeltaLossConfig.remove_reduced_attrs(data)


class ClippedDeltaLossConfig(BaseLossConfig):
    """Computes clipped delta Loss for DQN.

    [Wikipedia link](https://en.wikipedia.org/wiki/Huber_loss)
    [DeepMind link](https://sites.google.com/a/deepmind.com/dqn/)

    Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.
        clip_value_min: default to -1.
        clip_value_max: default to 1.

    Returns:
        A scalar `Tensor` representing the loss value.

    Raises:
        ValueError: If `predictions` shape doesn't match `labels` shape, or `weights` is `None`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: ClippedDeltaLoss
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      loss:
        ClippedDeltaLoss:
          clip_value_min: -0.8
      # other model properties
    ```
    """
    IDENTIFIER = 'ClippedDeltaLoss'
    SCHEMA = ClippedDeltaLossSchema

    def __init__(self,
                 input_layer=None,
                 output_layer=None,
                 weights=1.0,
                 clip_value_min=-1.,
                 clip_value_max=1.,
                 name=None,
                 collect=True):
        super(ClippedDeltaLossConfig, self).__init__(
            input_layer, output_layer, weights, name, collect)
        self.clip_value_min = clip_value_min
        self.clip_value_max = clip_value_max


class SoftmaxCrossEntropySchema(BaseLossSchema):
    label_smoothing = fields.Float(default=0., missing=0.)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return SoftmaxCrossEntropyConfig(**data)

    @post_dump
    def unmake(self, data):
        return SoftmaxCrossEntropyConfig.remove_reduced_attrs(data)


class SoftmaxCrossEntropyConfig(BaseLossConfig):
    """Computes Softmax Cross entropy (softmax categorical cross entropy).

    Computes softmax cross entropy between y_pred (logits) and
    y_true (labels).

    Measures the probability error in discrete classification tasks in which
    the classes are mutually exclusive (each entry is in exactly one class).
    For example, each CIFAR-10 image is labeled with one and only one label:
    an image can be a dog or a truck, but not both.

    **WARNING:** This op expects unscaled logits, since it performs a `softmax`
    on `y_pred` internally for efficiency.  Do not call this op with the
    output of `softmax`, as it will produce incorrect results.

    `y_pred` and `y_true` must have the same shape `[batch_size, num_classes]`
    and the same dtype (either `float32` or `float64`). It is also required
    that `y_true` (labels) are binary arrays (For example, class 2 out of a
    total of 5 different classes, will be define as [0., 1., 0., 0., 0.])

    Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.
        label_smoothing: If greater than `0` then smooth the labels.

    Returns:
        A scalar `Tensor` representing the loss value.

    Raises:
        ValueError: If `predictions` shape doesn't match `labels` shape, or `weights` is `None`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: SoftmaxCrossEntropy
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      loss:
        SoftmaxCrossEntropy:
          label_smoothing: 0.1
      # other model properties
    ```
    """
    IDENTIFIER = 'SoftmaxCrossEntropy'
    SCHEMA = SoftmaxCrossEntropySchema

    def __init__(self,
                 input_layer=None,
                 output_layer=None,
                 weights=1.0,
                 label_smoothing=0.,
                 name=None,
                 collect=True):
        super(SoftmaxCrossEntropyConfig, self).__init__(
            input_layer, output_layer, weights, name, collect)
        self.label_smoothing = label_smoothing


class SigmoidCrossEntropySchema(BaseLossSchema):
    label_smoothing = fields.Float(default=0., missing=0.)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return SigmoidCrossEntropyConfig(**data)

    @post_dump
    def unmake(self, data):
        return SigmoidCrossEntropyConfig.remove_reduced_attrs(data)


class SigmoidCrossEntropyConfig(BaseLossConfig):
    """Computes Sigmoid cross entropy.(binary cross entropy):

    Computes sigmoid cross entropy between y_pred (logits) and y_true
    (labels).

    Measures the probability error in discrete classification tasks in which
    each class is independent and not mutually exclusive. For instance,
    one could perform multilabel classification where a picture can contain
    both an elephant and a dog at the same time.

    For brevity, let `x = logits`, `z = targets`.  The logistic loss is

      x - x * z + log(1 + exp(-x))

    To ensure stability and avoid overflow, the implementation uses

      max(x, 0) - x * z + log(1 + exp(-abs(x)))

    `y_pred` and `y_true` must have the same type and shape.

    Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.
        label_smoothing: If greater than `0` then smooth the labels.

    Returns:
        A scalar `Tensor` representing the loss value.

    Raises:
        ValueError: If `predictions` shape doesn't match `labels` shape, or `weights` is `None`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: SigmoidCrossEntropy
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      loss:
        SigmoidCrossEntropy:
          label_smoothing: 0.1
      # other model properties
    ```
    """
    IDENTIFIER = 'SigmoidCrossEntropy'
    SCHEMA = SigmoidCrossEntropySchema

    def __init__(self,
                 input_layer=None,
                 output_layer=None,
                 weights=1.0,
                 label_smoothing=0.,
                 name=None,
                 collect=True):
        super(SigmoidCrossEntropyConfig, self).__init__(
            input_layer, output_layer, weights, name, collect)
        self.label_smoothing = label_smoothing


class HingeLossSchema(BaseLossSchema):
    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return HingeLossConfig(**data)

    @post_dump
    def unmake(self, data):
        return HingeLossConfig.remove_reduced_attrs(data)


class HingeLossConfig(BaseLossConfig):
    """Hinge Loss.

    Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.

    Returns:
        A scalar `Tensor` representing the loss value.

    Raises:
        ValueError: If `predictions` shape doesn't match `labels` shape, or `weights` is `None`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: HingeLoss
      # other model properties
    ```
    """
    IDENTIFIER = 'HingeLoss'
    SCHEMA = HingeLossSchema


class CosineDistanceSchema(BaseLossSchema):
    dim = fields.Int()

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return CosineDistanceConfig(**data)

    @post_dump
    def unmake(self, data):
        return CosineDistanceConfig.remove_reduced_attrs(data)


class CosineDistanceConfig(BaseLossConfig):
    """Adds a cosine-distance loss to the training procedure.

    Note that the function assumes that `predictions` and `labels` are already unit-normalized.

    WARNING: `weights` also supports dimensions of 1, but the broadcasting does
    not work as advertised, you'll wind up with weighted sum instead of weighted
    mean for any but the last dimension. This will be cleaned up soon, so please
    do not rely on the current behavior for anything but the shapes documented for
    `weights` below.

    Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.

    Returns:
        A scalar `Tensor` representing the loss value.

    Raises:
        ValueError: If `predictions` shape doesn't match `labels` shape, or `weights` is `None`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: CosineDistance
      # other model properties
    ```
    """
    IDENTIFIER = 'CosineDistance'
    SCHEMA = CosineDistanceSchema

    def __init__(self,
                 dim,
                 input_layer=None,
                 output_layer=None,
                 weights=1.0,
                 name=None,
                 collect=True):
        self.dim = dim
        super(CosineDistanceConfig, self).__init__(
            input_layer, output_layer, weights, name, collect)


class PoissonLossSchema(BaseLossSchema):
    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return PoissonLossConfig(**data)

    @post_dump
    def unmake(self, data):
        return PoissonLossConfig.remove_reduced_attrs(data)


class PoissonLossConfig(BaseLossConfig):
    """Adds a poisson loss to the training procedure.

    Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.

    Returns:
        A scalar `Tensor` representing the loss value.

    Raises:
        ValueError: If `predictions` shape doesn't match `labels` shape, or `weights` is `None`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: PoissonLoss
      # other model properties
    ```
    """
    IDENTIFIER = 'PoissonLoss'
    SCHEMA = PoissonLossSchema


class KullbackLeiberDivergenceSchema(BaseLossSchema):
    dim = fields.Int()

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return KullbackLeiberDivergenceConfig(**data)

    @post_dump
    def unmake(self, data):
        return KullbackLeiberDivergenceConfig.remove_reduced_attrs(data)


class KullbackLeiberDivergenceConfig(BaseLossConfig):
    """Adds a Kullback leiber diverenge loss to the training procedure.

     Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.

    Returns:
        A scalar `Tensor` representing the loss value.

    Raises:
        ValueError: If `predictions` shape doesn't match `labels` shape, or `weights` is `None`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: KullbackLeiberDivergence
      # other model properties
    ```
    """
    IDENTIFIER = 'KullbackLeiberDivergence'
    SCHEMA = KullbackLeiberDivergenceSchema

    def __init__(self,
                 dim,
                 input_layer=None,
                 output_layer=None,
                 weights=1.0, name='KullbackLeiberDivergence',
                 collect=True):
        self.dim = dim
        super(KullbackLeiberDivergenceConfig, self).__init__(
            input_layer, output_layer, weights, name, collect)


class LossSchema(BaseMultiSchema):
    __multi_schema_name__ = 'loss'
    __configs__ = {
        AbsoluteDifferenceConfig.IDENTIFIER: AbsoluteDifferenceConfig,
        MeanSquaredErrorConfig.IDENTIFIER: MeanSquaredErrorConfig,
        LogLossConfig.IDENTIFIER: LogLossConfig,
        HuberLossConfig.IDENTIFIER: HuberLossConfig,
        ClippedDeltaLossConfig.IDENTIFIER: ClippedDeltaLossConfig,
        SoftmaxCrossEntropyConfig.IDENTIFIER: SoftmaxCrossEntropyConfig,
        SigmoidCrossEntropyConfig.IDENTIFIER: SigmoidCrossEntropyConfig,
        HingeLossConfig.IDENTIFIER: HingeLossConfig,
        CosineDistanceConfig.IDENTIFIER: CosineDistanceConfig,
        KullbackLeiberDivergenceConfig.IDENTIFIER: KullbackLeiberDivergenceConfig,
    }



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import Schema, fields, post_dump, post_load

from polyaxon_schemas.base import BaseConfig, BaseMultiSchema


class BaseEnvironmentSchema(Schema):
    env_id = fields.Str()


class BaseEnvironmentConfig(BaseConfig):
    def __init__(self, env_id):
        self.env_id = env_id


class GymEnvironmentSchema(BaseEnvironmentSchema):
    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return GymEnvironmentConfig(**data)

    @post_dump
    def unmake(self, data):
        return GymEnvironmentConfig.remove_reduced_attrs(data)


class GymEnvironmentConfig(BaseEnvironmentConfig):
    IDENTIFIER = 'Gym'
    SCHEMA = GymEnvironmentSchema


class RegularizerSchema(BaseMultiSchema):
    __multi_schema_name__ = 'environment'
    __configs__ = {
        GymEnvironmentConfig.IDENTIFIER: GymEnvironmentConfig,
    }



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import abc
import six

import rhea

from hestia.list_utils import to_list
from marshmallow import ValidationError

from polyaxon_schemas.exceptions import PolyaxonConfigurationError, PolyaxonfileError
from polyaxon_schemas.operators import ForConfig, IfConfig
from polyaxon_schemas.polyaxonfile import validator
from polyaxon_schemas.polyaxonfile.parser import Parser
from polyaxon_schemas.polyaxonfile.utils import cached_property


@six.add_metaclass(abc.ABCMeta)
class BaseSpecification(object):
    """Base abstract specification for plyaxonfiles and configurations."""

    _EXPERIMENT = 'experiment'
    _GROUP = 'group'
    _JOB = 'job'
    _NOTEBOOK = 'notebook'
    _TENSORBOARD = 'tensorboard'
    _BUILD = 'build'
    _KINDS = {_EXPERIMENT, _GROUP, _JOB, _NOTEBOOK, _TENSORBOARD, _BUILD}

    _SPEC_KIND = None

    MAX_VERSION = 1.0  # Min Polyaxonfile specification version this CLI supports
    MIN_VERSION = 1.0  # Max Polyaxonfile specification version this CLI supports

    VERSION = 'version'
    KIND = 'kind'
    LOGGING = 'logging'
    TAGS = 'tags'
    HP_TUNING = 'hptuning'
    DECLARATIONS = 'declarations'
    ENVIRONMENT = 'environment'
    RUN = 'run'
    BUILD = 'build'
    MODEL = 'model'
    TRAIN = 'train'
    EVAL = 'eval'

    SECTIONS = (
        VERSION, KIND, TAGS, ENVIRONMENT, DECLARATIONS, LOGGING, HP_TUNING,
        BUILD, RUN, MODEL, TRAIN, EVAL
    )

    HEADER_SECTIONS = (
        VERSION, KIND, LOGGING, TAGS
    )

    GRAPH_SECTIONS = (
        MODEL, TRAIN, EVAL
    )

    REQUIRED_SECTIONS = (
        VERSION, KIND
    )

    POSSIBLE_SECTIONS = (
        VERSION, KIND, LOGGING, TAGS
    )

    OPERATORS = {
        ForConfig.IDENTIFIER: ForConfig,
        IfConfig.IDENTIFIER: IfConfig,
    }

    def __init__(self, values):
        self._values = to_list(values)

        try:
            self._data = rhea.read(self._values)
        except rhea.RheaError as e:
            raise PolyaxonConfigurationError(e)
        self.check_data()
        headers = Parser.get_headers(spec=self, data=self._data)
        try:
            self._headers = validator.validate_headers(spec=self, data=headers)
        except ValidationError as e:
            raise PolyaxonConfigurationError(e)
        self._parsed_data = None
        self._validated_data = None
        self._set_parsed_data()
        self._extra_validation()

    def _extra_validation(self):
        pass

    def _set_parsed_data(self):
        parsed_data = Parser.parse(self, self._data, None)
        self._validated_data = validator.validate(spec=self, data=parsed_data)
        self._parsed_data = parsed_data

    @classmethod
    def check_version(cls, data):
        if cls.VERSION not in data:
            raise PolyaxonfileError("The Polyaxonfile `version` must be specified.")
        if not cls.MIN_VERSION <= data[cls.VERSION] <= cls.MAX_VERSION:
            raise PolyaxonfileError(
                "The Polyaxonfile's version specified is not supported by your current CLI."
                "Your CLI support Polyaxonfile versions between: {} {}."
                "You can run `polyaxon upgrade` and "
                "check documentation for the specification.".format(
                    cls.MIN_VERSION, cls.MAX_VERSION))

    @classmethod
    def check_kind(cls, data):
        if cls.KIND not in data:
            raise PolyaxonfileError("The Polyaxonfile `kind` must be specified.")

        if data[cls.KIND] not in cls._KINDS:
            raise PolyaxonfileError(
                "The Polyaxonfile with kind `{}` is not a supported value.".format(data[cls.KIND]))

    def check_data(self, data=None):
        data = data or self._data
        self.check_version(data)
        self.check_kind(data)
        if data[self.KIND] != self._SPEC_KIND:
            raise PolyaxonfileError(
                "The specification used `{}` is incompatible with the kind `{}`.".format(
                    self.__class__.__name__, data[self.KIND]))
        for key in set(six.iterkeys(data)) - set(self.SECTIONS):
            raise PolyaxonfileError(
                "Unexpected section `{}` in Polyaxonfile version `{}`. "
                "Please check the Polyaxonfile specification "
                "for this version.".format(key, data[self.VERSION]))

        for key in set(six.iterkeys(data)) - set(self.POSSIBLE_SECTIONS):
            raise PolyaxonfileError(
                "Unexpected section `{}` for specification kind `{}` version `{}`. "
                "Please check the Polyaxonfile specification "
                "for this version.".format(key, self._SPEC_KIND, data[self.VERSION]))

        for key in self.REQUIRED_SECTIONS:
            if key not in data:
                raise PolyaxonfileError("{} is a required section for a valid Polyaxonfile".format(
                    key))

    def patch(self, values):
        values = [self._parsed_data] + to_list(values)
        return self.read(values=values)

    @classmethod
    def get_kind(cls, data):
        cls.check_kind(data=data)
        return data[cls.KIND]

    @classmethod
    def read(cls, values):
        if isinstance(values, cls):
            return values
        return cls(values)

    @cached_property
    def is_experiment(self):
        return self.kind == self._EXPERIMENT

    @cached_property
    def is_group(self):
        return self.kind == self._GROUP

    @cached_property
    def is_job(self):
        return self.kind == self._JOB

    @cached_property
    def is_notebook(self):
        return self.kind == self._NOTEBOOK

    @cached_property
    def is_tensorboard(self):
        return self.kind == self._TENSORBOARD

    @cached_property
    def is_build(self):
        return self.kind == self._BUILD

    @property
    def values(self):
        return self._values

    @cached_property
    def data(self):
        return self._data

    @cached_property
    def headers(self):
        return self._headers

    @cached_property
    def parsed_data(self):
        return self._parsed_data

    @cached_property
    def version(self):
        return self.headers[self.VERSION]

    @cached_property
    def kind(self):
        return self.headers[self.KIND]

    @cached_property
    def logging(self):
        return self.headers.get(self.LOGGING, None)

    @cached_property
    def log_level(self):
        if self.logging:
            return self.logging.level
        return 'INFO'

    @cached_property
    def tags(self):
        tags = self.headers.get(self.TAGS, None)
        return list(set(tags)) if tags else None



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import os

import rhea

from hestia.list_utils import to_list

from polyaxon_schemas.exceptions import PolyaxonConfigurationError, PolyaxonfileError
from polyaxon_schemas.polyaxonfile.specification import SPECIFICATION_BY_KIND
from polyaxon_schemas.polyaxonfile.specification.base import BaseSpecification


class PolyaxonFile(object):
    """Parses Polyaxonfiles, and validate that it respects the current file specification"""

    def __init__(self, filepaths):
        filepaths = to_list(filepaths)
        for filepath in filepaths:
            if not os.path.isfile(filepath):
                raise PolyaxonfileError("`{}` must be a valid file".format(filepath))
        self._filenames = [os.path.basename(filepath) for filepath in filepaths]
        data = rhea.read(filepaths)
        kind = BaseSpecification.get_kind(data=data)
        try:
            self.specification = SPECIFICATION_BY_KIND[kind](data)
        except PolyaxonConfigurationError as e:
            raise PolyaxonfileError(e)

    @property
    def filenames(self):
        return self._filenames



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import Schema, fields, post_dump, post_load

from polyaxon_schemas.base import BaseConfig, BaseMultiSchema


class BaseMemorySchema(Schema):
    size = fields.Int(allow_none=True)
    batch_size = fields.Int(allow_none=True)

    @post_load
    def make(self, data):
        return BaseMemoryConfig(**data)

    @post_dump
    def unmake(self, data):
        return BaseMemoryConfig.remove_reduced_attrs(data)


class BaseMemoryConfig(BaseConfig):
    IDENTIFIER = 'Memory'
    SCHEMA = BaseMemorySchema

    def __init__(self, size=5000, batch_size=5000):
        self.size = size
        self.batch_size = batch_size


class BatchMemorySchema(BaseMemorySchema):
    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return BatchMemoryConfig(**data)

    @post_dump
    def unmake(self, data):
        return BatchMemoryConfig.remove_reduced_attrs(data)


class BatchMemoryConfig(BaseMemoryConfig):
    IDENTIFIER = 'BatchMemory'
    SCHEMA = BatchMemorySchema


class MemorySchema(BaseMultiSchema):
    __multi_schema_name__ = 'memory'
    __configs__ = {
        BaseMemoryConfig.IDENTIFIER: BaseMemoryConfig,
        BatchMemoryConfig.IDENTIFIER: BatchMemoryConfig,
    }



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_dict, assert_tensors

from polyaxon_schemas.ml.metrics import (
    AccuracyConfig,
    AUCConfig,
    CovarianceConfig,
    FalseNegativesConfig,
    FalsePositivesConfig,
    MeanAbsoluteErrorConfig,
    MeanConfig,
    MeanCosineDistanceConfig,
    MeanIOUConfig,
    MeanRelativeErrorConfig,
    MeanSquaredErrorConfig,
    MeanTensorConfig,
    PearsonCorrelationConfig,
    PercentageLessConfig,
    PrecisionAtThresholdsConfig,
    PrecisionConfig,
    RecallAtThresholdsConfig,
    RecallConfig,
    RootMeanSquaredErrorConfig,
    SensitivityAtSpecificityConfig,
    SparsePrecisionAtKConfig,
    SpecificityAtSensitivityConfig,
    TrueNegativesConfig,
    TruePositivesConfig
)


class TestMetricConfigs(TestCase):
    @staticmethod
    def assert_equal_metrics(m1, m2):
        assert_tensors(m1.pop('input_layer', None), m2.pop('input_layer', None))
        assert_tensors(m1.pop('output_layer', None), m2.pop('output_layer', None))

        assert_equal_dict(m1, m2)

    def test_base_metrics_config(self):
        config_dict = {
            'input_layer': 'images',
            'output_layer': 'relu_1',
            'weights': None,
            'name': 'm'
        }

        config_classes = [
            TruePositivesConfig,
            TrueNegativesConfig,
            FalsePositivesConfig,
            FalseNegativesConfig,
            AccuracyConfig,
            PrecisionConfig,
            RecallConfig,
            MeanAbsoluteErrorConfig,
            MeanSquaredErrorConfig,
            RootMeanSquaredErrorConfig,
            CovarianceConfig,
            PearsonCorrelationConfig,
        ]

        for config_class in config_classes:
            config = config_class.from_dict(config_dict)
            self.assert_equal_metrics(config.to_dict(), config_dict)

    def test_auc_metric_config(self):
        config_dict = {
            'input_layer': 'images',
            'output_layer': 'relu_1',
            'specificity': 0.1,
            'num_thresholds': 300,
            'weights': None,
            'curve': 'ROC',
            'name': 'm',
        }
        config = AUCConfig.from_dict(config_dict)
        self.assert_equal_metrics(config.to_dict(), config_dict)

    def test_mean_metric_config(self):
        config_dicts = [
            {
                'values': ['images', 0, 0],
                'weights': None,
                'name': 'm'
            },
            {
                'values': ['images', 0, 0],
                'weights': None,
                'name': 'm'
            }
        ]

        for config_dict in config_dicts:
            config = MeanConfig.from_dict(config_dict)
            self.assert_equal_metrics(config.to_dict(), config_dict)

    def test_mean_tensor_metric_config(self):
        config_dicts = [
            {
                'values': ['images', 0, 0],
                'weights': None,
                'name': 'm'
            },
            {
                'values': ['images', 0, 0],
                'weights': None,
                'name': 'm'
            }
        ]

        for config_dict in config_dicts:
            config = MeanTensorConfig.from_dict(config_dict)
            self.assert_equal_metrics(config.to_dict(), config_dict)

    def test_specificity_at_sensitivity_metric_config(self):
        config_dict = {
            'input_layer': 'images',
            'output_layer': 'relu_1',
            'sensitivity': 0.1,
            'num_thresholds': 300,
            'weights': None,
            'name': 'm'
        }
        config = SpecificityAtSensitivityConfig.from_dict(config_dict)
        self.assert_equal_metrics(config.to_dict(), config_dict)

    def test_sensitivity_at_specificity_metric_config(self):
        config_dict = {
            'input_layer': 'images',
            'output_layer': 'relu_1',
            'specificity': 0.1,
            'num_thresholds': 300,
            'weights': None,
            'name': 'm'
        }
        config = SensitivityAtSpecificityConfig.from_dict(config_dict)
        self.assert_equal_metrics(config.to_dict(), config_dict)

    def test_metric_at_thresholds_config(self):
        config_dict = {
            'input_layer': 'images',
            'output_layer': 'relu_1',
            'thresholds': [0.1, 0.1, 0.1],
            'weights': None,
            'name': 'm'
        }
        config_classes = [PrecisionAtThresholdsConfig, RecallAtThresholdsConfig]
        for config_class in config_classes:
            config = config_class.from_dict(config_dict)
            self.assert_equal_metrics(config.to_dict(), config_dict)

    def test_sparse_precistion_at_k_metric_config(self):
        config_dict = {
            'input_layer': 'images',
            'output_layer': 'relu_1',
            'k': 1,
            'class_id': None,
            'weights': None,
            'name': 'm'
        }
        config = SparsePrecisionAtKConfig.from_dict(config_dict)
        self.assert_equal_metrics(config.to_dict(), config_dict)

    def test_mean_relative_error_config(self):
        config_dict = {
            'input_layer': 'images',
            'output_layer': 'relu_1',
            'normalizer': 'tensor_normalizer',
            'weights': None,
            'name': 'm'
        }
        config = MeanRelativeErrorConfig.from_dict(config_dict)
        self.assert_equal_metrics(config.to_dict(), config_dict)

    def test_mean_consine_distance_metric_config(self):
        config_dict = {
            'input_layer': 'images',
            'output_layer': 'relu_1',
            'dim': 1,
            'weights': None,
            'name': 'm'
        }
        config = MeanCosineDistanceConfig.from_dict(config_dict)
        self.assert_equal_metrics(config.to_dict(), config_dict)

    def test_percentage_less_metric_config(self):
        config_dict = {
            'values': ['images', 0, 0],
            'threshold': 0.4,
            'weights': None,
            'name': 'm'
        }
        config = PercentageLessConfig.from_dict(config_dict)
        self.assert_equal_metrics(config.to_dict(), config_dict)

    def test_mean_iou_config(self):
        config_dict = {
            'input_layer': 'images',
            'output_layer': 'relu_1',
            'num_classes': 4,
            'weights': None,
            'name': 'm'
        }
        config = MeanIOUConfig.from_dict(config_dict)
        self.assert_equal_metrics(config.to_dict(), config_dict)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from polyaxon_schemas.code_reference import CodeReferenceConfig


class TestCodeReferenceConfigs(TestCase):
    def test_code_reference_config(self):
        config_dict = {
            'id': 1,
            'commit': '3783ab36703b14b91b15736fe4302bfb8d52af1c',
            'head': '3783ab36703b14b91b15736fe4302bfb8d52af1c',
            'branch': 'feature1',
            'git_url': 'https://bitbucket.org:foo/bar.git',
            'is_dirty': True
        }
        config = CodeReferenceConfig.from_dict(config_dict)
        config_to_dict = config.to_dict()
        assert config_to_dict == config_dict

        config_dict = {
            'commit': '3783ab36703b14b91b15736fe4302bfb8d52af1c',
            'head': '3783ab36703b14b91b15736fe4302bfb8d52af1c',
            'branch': 'feature1',
            'git_url': 'https://bitbucket.org:foo/bar.git',
            'is_dirty': True
        }
        config = CodeReferenceConfig.from_dict(config_dict)
        assert config.to_light_dict() == config_dict



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function


class PolyaxonSchemaError(Exception):
    """Exception class to raise if a schema has an issue."""
    pass


class PolyaxonfileError(Exception):
    """Exception class to raise if an error is encountered during the parsing of a Polyaxonfile."""
    pass


class PolyaxonConfigurationError(Exception):
    """Exception class to raise if a Configurable object has an issue."""
    pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_layers

from polyaxon_schemas.ml.constraints import MaxNormConfig
from polyaxon_schemas.ml.initializations import (
    GlorotNormalInitializerConfig,
    ZerosInitializerConfig
)
from polyaxon_schemas.ml.layers.convolutional import (
    Conv1DConfig,
    Conv2DConfig,
    Conv2DTransposeConfig,
    Conv3DConfig,
    Conv3DTransposeConfig,
    Cropping1DConfig,
    Cropping2DConfig,
    Cropping3DConfig,
    SeparableConv2DConfig,
    UpSampling1DConfig,
    UpSampling2DConfig,
    UpSampling3DConfig,
    ZeroPadding1DConfig,
    ZeroPadding2DConfig,
    ZeroPadding3DConfig
)
from polyaxon_schemas.ml.regularizations import L1L2RegularizerConfig, L1RegularizerConfig


class TestConvolutionConfigs(TestCase):
    @staticmethod
    def assert_conv_config(conv_class, dim):
        config_dict = {
            'filters': 30,
            'kernel_size': 3,
            'strides': 1 if dim == 1 else [1, 1] if dim == 2 else [1, 1, 1],
            'padding': 'valid',
            'activation': 'relu',
            'dilation_rate': 1 if dim == 1 else [1, 1] if dim == 2 else [1, 1, 1],
            'use_bias': True,
            'kernel_initializer': GlorotNormalInitializerConfig().to_schema(),
            'bias_initializer': ZerosInitializerConfig().to_schema(),
            'kernel_regularizer': L1L2RegularizerConfig().to_schema(),
            'bias_regularizer': None,
            'activity_regularizer': L1RegularizerConfig().to_schema(),
            'kernel_constraint': MaxNormConfig().to_schema(),
            'bias_constraint': None,
            'inbound_nodes': [['layer_1', 0, 1], ['layer_2', 1, 1]]
        }
        if dim > 1:
            config_dict['data_format'] = None
        config = conv_class.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    @staticmethod
    def assert_separable_conv(conv_class, dim):
        config_dict = {
            'filters': 30,
            'kernel_size': 10,
            'strides': [1, 1] if dim == 2 else [1, 1, 1],
            'padding': 'valid',
            'data_format': None,
            'depth_multiplier': 1,
            'activation': None,
            'use_bias': True,
            'depthwise_initializer': GlorotNormalInitializerConfig().to_schema(),
            'pointwise_initializer': GlorotNormalInitializerConfig().to_schema(),
            'bias_initializer': ZerosInitializerConfig().to_schema(),
            'depthwise_regularizer': None,
            'pointwise_regularizer': None,
            'bias_regularizer': None,
            'activity_regularizer': None,
            'depthwise_constraint': None,
            'pointwise_constraint': None,
            'bias_constraint': None
        }
        config = conv_class.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    @staticmethod
    def assert_upsampling(upsampling_class, dim):
        config_dict = {
            'size': 1 if dim == 1 else [1, 1] if dim == 2 else [1, 1, 1],
        }
        if dim > 1:
            config_dict['data_format'] = None
        config = upsampling_class.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    @staticmethod
    def assert_zero_padding(zero_padding_class, dim):
        config_dict = {
            'padding': 1 if dim == 1 else [1, 1] if dim == 2 else [1, 1, 1],
        }
        if dim > 1:
            config_dict['data_format'] = None
        config = zero_padding_class.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    @staticmethod
    def assert_cropping(zero_padding_class, dim):
        config_dict = {
            'cropping': (
                1 if dim == 1 else [[1, 1], [1, 1]]
                if dim == 2 else [[1, 1], [1, 1], [1, 1]]
            ),
        }
        if dim > 1:
            config_dict['data_format'] = None
        config = zero_padding_class.from_dict(config_dict)
        assert_equal_layers(config, config_dict)

    def test_conv_config(self):
        self.assert_conv_config(Conv1DConfig, 1)
        self.assert_conv_config(Conv2DConfig, 2)
        self.assert_conv_config(Conv3DConfig, 3)
        self.assert_conv_config(Conv2DTransposeConfig, 2)
        self.assert_conv_config(Conv3DTransposeConfig, 3)

    def test_separable_conv_config(self):
        self.assert_separable_conv(SeparableConv2DConfig, 2)

    def test_upsampling_config(self):
        self.assert_upsampling(UpSampling1DConfig, 1)
        self.assert_upsampling(UpSampling2DConfig, 2)
        self.assert_upsampling(UpSampling3DConfig, 3)

    def test_zero_padding_config(self):
        self.assert_zero_padding(ZeroPadding1DConfig, 1)
        self.assert_zero_padding(ZeroPadding2DConfig, 2)
        self.assert_zero_padding(ZeroPadding3DConfig, 3)

    def test_cropping_config(self):
        self.assert_cropping(Cropping1DConfig, 1)
        self.assert_cropping(Cropping2DConfig, 2)
        self.assert_cropping(Cropping3DConfig, 3)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, post_dump, post_load, validate

from polyaxon_schemas.ml.constraints import ConstraintSchema
from polyaxon_schemas.ml.initializations import (
    GlorotUniformInitializerConfig,
    InitializerSchema,
    OrthogonalInitializerConfig,
    ZerosInitializerConfig
)
from polyaxon_schemas.ml.layers.base import BaseLayerConfig, BaseLayerSchema
from polyaxon_schemas.ml.regularizations import RegularizerSchema
from polyaxon_schemas.utils import ACTIVATION_VALUES, StrOrFct


class RecurrentSchema(BaseLayerSchema):
    return_sequences = fields.Bool(default=False, missing=False)
    return_state = fields.Bool(default=False, missing=False)
    go_backwards = fields.Bool(default=False, missing=False)
    stateful = fields.Bool(default=False, missing=False)
    unroll = fields.Bool(default=False, missing=False)
    implementation = fields.Int(default=0, missing=0)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return RecurrentConfig(**data)

    @post_dump
    def unmake(self, data):
        return RecurrentConfig.remove_reduced_attrs(data)


class RecurrentConfig(BaseLayerConfig):
    """Abstract base class for recurrent layers.

    Do not use in a model -- it's not a valid layer!
    Use its children classes `LSTM`, `GRU` and `SimpleRNN` instead.

    All recurrent layers (`LSTM`, `GRU`, `SimpleRNN`) also
    follow the specifications of this class and accept
    the keyword arguments listed below.

    Example:

    ```python
        # as the first layer in a Sequential model
        model = Sequential()
        model.add(LSTM(32, input_shape=(10, 64)))
        # now model.output_shape == (None, 32)
        # note: `None` is the batch dimension.

        # for subsequent layers, no need to specify the input size:
        model.add(LSTM(16))

        # to stack recurrent layers, you must use return_sequences=True
        # on any recurrent layer that feeds into another recurrent layer.
        # note that you only need to specify the input size on the first layer.
        model = Sequential()
        model.add(LSTM(64, input_dim=64, input_length=10, return_sequences=True))
        model.add(LSTM(32, return_sequences=True))
        model.add(LSTM(10))
    ```

    Args:
        weights: list of Numpy arrays to set as initial weights.
            The list should have 3 elements, of shapes:
            `[(input_dim, output_dim), (output_dim, output_dim), (output_dim,)]`.
        return_sequences: Boolean. Whether to return the last output
            in the output sequence, or the full sequence.
        return_state: Boolean. Whether to return the last state
            in addition to the output.
        go_backwards: Boolean (default False).
            If True, process the input sequence backwards and return the
            reversed sequence.
        stateful: Boolean (default False). If True, the last state
            for each sample at index i in a batch will be used as initial
            state for the sample of index i in the following batch.
        unroll: Boolean (default False).
            If True, the network will be unrolled,
            else a symbolic loop will be used.
            Unrolling can speed-up a RNN,
            although it tends to be more memory-intensive.
            Unrolling is only suitable for short sequences.
        implementation: one of {0, 1, or 2}.
            If set to 0, the RNN will use
            an implementation that uses fewer, larger matrix products,
            thus running faster on CPU but consuming more memory.
            If set to 1, the RNN will use more matrix products,
            but smaller ones, thus running slower
            (may actually be faster on GPU) while consuming less memory.
            If set to 2 (LSTM/GRU only),
            the RNN will combine the input gate,
            the forget gate and the output gate into a single matrix,
            enabling more time-efficient parallelization on the GPU.
            Note: RNN dropout must be shared for all gates,
            resulting in a slightly reduced regularization.
        input_dim: dimensionality of the input (integer).
            This argument (or alternatively, the keyword argument `input_shape`)
            is required when using this layer as the first layer in a model.
        input_length: Length of input sequences, to be specified
            when it is constant.
            This argument is required if you are going to connect
            `Flatten` then `Dense` layers upstream
            (without it, the shape of the dense outputs cannot be computed).
            Note that if the recurrent layer is not the first layer
            in your model, you would need to specify the input length
            at the level of the first layer
            (e.g. via the `input_shape` argument)

    Input shape:s
        3D tensor with shape `(batch_size, timesteps, input_dim)`,
        (Optional) 2D tensors with shape `(batch_size, output_dim)`.

    Output shape:
        - if `return_state`: a list of tensors. The first tensor is
            the output. The remaining tensors are the last states,
            each with shape `(batch_size, units)`.
        - if `return_sequences`: 3D tensor with shape
            `(batch_size, timesteps, units)`.
        - else, 2D tensor with shape `(batch_size, units)`.

    # Masking
        This layer supports masking for input data with a variable number
        of timesteps. To introduce masks to your data,
        use an `Embedding` layer with the `mask_zero` parameter
        set to `True`.

    # Note on using statefulness in RNNs
        You can set RNN layers to be 'stateful', which means that the states
        computed for the samples in one batch will be reused as initial states
        for the samples in the next batch. This assumes a one-to-one mapping
        between samples in different successive batches.

        To enable statefulness:
            - specify `stateful=True` in the layer constructor.
            - specify a fixed batch size for your model, by passing
                if sequential model:
                  `batch_input_shape=(...)` to the first layer in your model.
                else for functional model with 1 or more Input layers:
                  `batch_shape=(...)` to all the first layers in your model.
                This is the expected shape of your inputs
                *including the batch size*.
                It should be a tuple of integers, e.g. `(32, 10, 100)`.
            - specify `shuffle=False` when calling fit().

        To reset the states of your model, call `.reset_states()` on either
        a specific layer, or on your entire model.

    # Note on specifying the initial state of RNNs
        You can specify the initial state of RNN layers symbolically by
        calling them with the keyword argument `initial_state`. The value of
        `initial_state` should be a tensor or list of tensors representing
        the initial state of the RNN layer.

        You can specify the initial state of RNN layers numerically by
        calling `reset_states` with the keyword argument `states`. The value of
        `states` should be a numpy array or list of numpy arrays representing
        the initial state of the RNN layer.
    """
    IDENTIFIER = 'Recurrent'
    SCHEMA = RecurrentSchema

    def __init__(self,
                 return_sequences=False,
                 return_state=False,
                 go_backwards=False,
                 stateful=False,
                 unroll=False,
                 implementation=0,
                 **kwargs):
        super(RecurrentConfig, self).__init__(**kwargs)
        self.return_sequences = return_sequences
        self.return_state = return_state
        self.go_backwards = go_backwards
        self.stateful = stateful
        self.unroll = unroll
        self.implementation = implementation


class SimpleRNNSchema(RecurrentSchema):
    units = fields.Int()
    activation = StrOrFct(allow_none=True, validate=validate.OneOf(ACTIVATION_VALUES))
    use_bias = fields.Bool(default=True, missing=True)
    kernel_initializer = fields.Nested(InitializerSchema, default=None, missing=None)
    recurrent_initializer = fields.Nested(InitializerSchema, default=None, missing=None)
    bias_initializer = fields.Nested(InitializerSchema, default=None, missing=None)
    kernel_regularizer = fields.Nested(RegularizerSchema, default=None, missing=None)
    recurrent_regularizer = fields.Nested(RegularizerSchema, default=None, missing=None)
    bias_regularizer = fields.Nested(RegularizerSchema, default=None, missing=None)
    activity_regularizer = fields.Nested(RegularizerSchema, default=None, missing=None)
    kernel_constraint = fields.Nested(ConstraintSchema, default=None, missing=None)
    recurrent_constraint = fields.Nested(ConstraintSchema, default=None, missing=None)
    bias_constraint = fields.Nested(ConstraintSchema, default=None, missing=None)
    dropout = fields.Float(default=0., missing=0.)
    recurrent_dropout = fields.Float(default=0., missing=0.)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return SimpleRNNConfig(**data)

    @post_dump
    def unmake(self, data):
        return SimpleRNNConfig.remove_reduced_attrs(data)


class SimpleRNNConfig(RecurrentConfig):
    """Fully-connected RNN where the output is to be fed back to input.

    Args:
        units: Positive integer, dimensionality of the output space.
        activation: Activation function to use.
            If you don't specify anything, no activation is applied
            If you pass None, no activation is applied
            (ie. "linear" activation: `a(x) = x`).
        use_bias: Boolean, whether the layer uses a bias vector.
        kernel_initializer: Initializer for the `kernel` weights matrix,
            used for the linear transformation of the inputs..
        recurrent_initializer: Initializer for the `recurrent_kernel`
            weights matrix,
            used for the linear transformation of the recurrent state..
        bias_initializer: Initializer for the bias vector.
        kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix.
        recurrent_regularizer: Regularizer function applied to
            the `recurrent_kernel` weights matrix.
        bias_regularizer: Regularizer function applied to the bias vector.
        activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation")..
        kernel_constraint: Constraint function applied to
            the `kernel` weights matrix.
        recurrent_constraint: Constraint function applied to
            the `recurrent_kernel` weights matrix.
        bias_constraint: Constraint function applied to the bias vector.
        dropout: Float between 0 and 1.
            Fraction of the units to drop for
            the linear transformation of the inputs.
        recurrent_dropout: Float between 0 and 1.
            Fraction of the units to drop for
            the linear transformation of the recurrent state.

    References:
        - [A Theoretically Grounded Application of Dropout in Recurrent Neural
          Networks](http://arxiv.org/abs/1512.05287)

    Polyaxonfile usage:

    ```yaml
    SimpleRNN:
      units: 3
      activation: tanh
    ```
    """
    IDENTIFIER = 'SimpleRNN'
    SCHEMA = SimpleRNNSchema

    def __init__(self,
                 units,
                 activation='tanh',
                 use_bias=True,
                 kernel_initializer=GlorotUniformInitializerConfig(),
                 recurrent_initializer=OrthogonalInitializerConfig(),
                 bias_initializer=ZerosInitializerConfig(),
                 kernel_regularizer=None,
                 recurrent_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 recurrent_constraint=None,
                 bias_constraint=None,
                 dropout=0.,
                 recurrent_dropout=0.,
                 **kwargs):
        super(SimpleRNNConfig, self).__init__(**kwargs)
        self.units = units
        self.activation = activation
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.recurrent_initializer = recurrent_initializer
        self.bias_initializer = bias_initializer
        self.kernel_regularizer = kernel_regularizer
        self.recurrent_regularizer = recurrent_regularizer
        self.bias_regularizer = bias_regularizer
        self.activity_regularizer = activity_regularizer
        self.kernel_constraint = kernel_constraint
        self.recurrent_constraint = recurrent_constraint
        self.bias_constraint = bias_constraint
        self.dropout = dropout
        self.recurrent_dropout = recurrent_dropout


class GRUSchema(RecurrentSchema):
    units = fields.Int()
    activation = StrOrFct(allow_none=True, validate=validate.OneOf(ACTIVATION_VALUES))
    use_bias = fields.Bool(default=True, missing=True)
    recurrent_activation = StrOrFct(allow_none=True, validate=validate.OneOf(ACTIVATION_VALUES))
    kernel_initializer = fields.Nested(InitializerSchema, default=None, missing=None)
    recurrent_initializer = fields.Nested(InitializerSchema, default=None, missing=None)
    bias_initializer = fields.Nested(InitializerSchema, default=None, missing=None)
    kernel_regularizer = fields.Nested(RegularizerSchema, default=None, missing=None)
    recurrent_regularizer = fields.Nested(RegularizerSchema, default=None, missing=None)
    bias_regularizer = fields.Nested(RegularizerSchema, default=None, missing=None)
    activity_regularizer = fields.Nested(RegularizerSchema, default=None, missing=None)
    kernel_constraint = fields.Nested(ConstraintSchema, default=None, missing=None)
    recurrent_constraint = fields.Nested(ConstraintSchema, default=None, missing=None)
    bias_constraint = fields.Nested(ConstraintSchema, default=None, missing=None)
    dropout = fields.Float(default=0., missing=0.)
    recurrent_dropout = fields.Float(default=0., missing=0.)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return GRUConfig(**data)

    @post_dump
    def unmake(self, data):
        return GRUConfig.remove_reduced_attrs(data)


class GRUConfig(RecurrentConfig):
    """Gated Recurrent Unit - Cho et al.

    2014.

    Args:
        units: Positive integer, dimensionality of the output space.
        activation: Activation function to use.
            If you pass None, no activation is applied
            (ie. "linear" activation: `a(x) = x`).
        recurrent_activation: Activation function to use
            for the recurrent step.
        use_bias: Boolean, whether the layer uses a bias vector.
        kernel_initializer: Initializer for the `kernel` weights matrix,
            used for the linear transformation of the inputs..
        recurrent_initializer: Initializer for the `recurrent_kernel`
            weights matrix,
            used for the linear transformation of the recurrent state..
        bias_initializer: Initializer for the bias vector.
        kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix.
        recurrent_regularizer: Regularizer function applied to
            the `recurrent_kernel` weights matrix.
        bias_regularizer: Regularizer function applied to the bias vector.
        activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation")..
        kernel_constraint: Constraint function applied to
            the `kernel` weights matrix.
        recurrent_constraint: Constraint function applied to
            the `recurrent_kernel` weights matrix.
        bias_constraint: Constraint function applied to the bias vector.
        dropout: Float between 0 and 1.
            Fraction of the units to drop for
            the linear transformation of the inputs.
        recurrent_dropout: Float between 0 and 1.
            Fraction of the units to drop for
            the linear transformation of the recurrent state.

    References:
        - [On the Properties of Neural Machine Translation: Encoder-Decoder
          Approaches](https://arxiv.org/abs/1409.1259)
        - [Empirical Evaluation of Gated Recurrent Neural Networks on Sequence
          Modeling](http://arxiv.org/abs/1412.3555v1)
        - [A Theoretically Grounded Application of Dropout in Recurrent Neural
          Networks](http://arxiv.org/abs/1512.05287)

    Polyaxonfile usage:

    ```yaml
    GRU:
      units: 3
      activation: tanh
    ```
    """
    IDENTIFIER = 'GRU'
    SCHEMA = GRUSchema

    def __init__(self,
                 units,
                 activation='tanh',
                 recurrent_activation='hard_sigmoid',
                 use_bias=True,
                 kernel_initializer=GlorotUniformInitializerConfig(),
                 recurrent_initializer=OrthogonalInitializerConfig(),
                 bias_initializer=ZerosInitializerConfig(),
                 kernel_regularizer=None,
                 recurrent_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 recurrent_constraint=None,
                 bias_constraint=None,
                 dropout=0.,
                 recurrent_dropout=0.,
                 **kwargs):
        super(GRUConfig, self).__init__(**kwargs)
        self.units = units
        self.activation = activation
        self.recurrent_activation = recurrent_activation
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.recurrent_initializer = recurrent_initializer
        self.bias_initializer = bias_initializer
        self.kernel_regularizer = kernel_regularizer
        self.recurrent_regularizer = recurrent_regularizer
        self.bias_regularizer = bias_regularizer
        self.activity_regularizer = activity_regularizer
        self.kernel_constraint = kernel_constraint
        self.recurrent_constraint = recurrent_constraint
        self.bias_constraint = bias_constraint
        self.dropout = dropout
        self.recurrent_dropout = recurrent_dropout


class LSTMSchema(RecurrentSchema):
    units = fields.Int()
    activation = StrOrFct(allow_none=True, validate=validate.OneOf(ACTIVATION_VALUES))
    use_bias = fields.Bool(default=True, missing=True)
    recurrent_activation = StrOrFct(allow_none=True, validate=validate.OneOf(ACTIVATION_VALUES))
    kernel_initializer = fields.Nested(InitializerSchema, default=None, missing=None)
    recurrent_initializer = fields.Nested(InitializerSchema, default=None, missing=None)
    bias_initializer = fields.Nested(InitializerSchema, default=None, missing=None)
    unit_forget_bias = fields.Bool(default=True, missing=True)
    kernel_regularizer = fields.Nested(RegularizerSchema, default=None, missing=None)
    recurrent_regularizer = fields.Nested(RegularizerSchema, default=None, missing=None)
    bias_regularizer = fields.Nested(RegularizerSchema, default=None, missing=None)
    activity_regularizer = fields.Nested(RegularizerSchema, default=None, missing=None)
    kernel_constraint = fields.Nested(ConstraintSchema, default=None, missing=None)
    recurrent_constraint = fields.Nested(ConstraintSchema, default=None, missing=None)
    bias_constraint = fields.Nested(ConstraintSchema, default=None, missing=None)
    dropout = fields.Float(default=0., missing=0.)
    recurrent_dropout = fields.Float(default=0., missing=0.)

    class Meta:
        ordered = True

    @post_load
    def make(self, data):
        return LSTMConfig(**data)

    @post_dump
    def unmake(self, data):
        return LSTMConfig.remove_reduced_attrs(data)


class LSTMConfig(RecurrentConfig):
    """Long-Short Term Memory unit - Hochreiter 1997.

    For a step-by-step description of the algorithm, see
    [this tutorial](http://deeplearning.net/tutorial/lstm.html).

    Args:
        units: Positive integer, dimensionality of the output space.
        activation: Activation function to use.
            If you pass None, no activation is applied
            (ie. "linear" activation: `a(x) = x`).
        recurrent_activation: Activation function to use
            for the recurrent step.
        use_bias: Boolean, whether the layer uses a bias vector.
        kernel_initializer: Initializer for the `kernel` weights matrix,
            used for the linear transformation of the inputs..
        recurrent_initializer: Initializer for the `recurrent_kernel`
            weights matrix,
            used for the linear transformation of the recurrent state..
        bias_initializer: Initializer for the bias vector.
        unit_forget_bias: Boolean.
            If True, add 1 to the bias of the forget gate at initialization.
            Setting it to true will also force `bias_initializer="zeros"`.
            This is recommended in [Jozefowicz et
              al.](http://www.jmlr.org/proceedings/papers/v37/jozefowicz15.pdf)
        kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix.
        recurrent_regularizer: Regularizer function applied to
            the `recurrent_kernel` weights matrix.
        bias_regularizer: Regularizer function applied to the bias vector.
        activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation")..
        kernel_constraint: Constraint function applied to
            the `kernel` weights matrix.
        recurrent_constraint: Constraint function applied to
            the `recurrent_kernel` weights matrix.
        bias_constraint: Constraint function applied to the bias vector.
        dropout: Float between 0 and 1.
            Fraction of the units to drop for
            the linear transformation of the inputs.
        recurrent_dropout: Float between 0 and 1.
            Fraction of the units to drop for
            the linear transformation of the recurrent state.

    References:
        - [Long short-term
          memory]((http://www.bioinf.jku.at/publications/older/2604.pdf)
          (original 1997 paper)
        - [Supervised sequence labeling with recurrent neural
          networks](http://www.cs.toronto.edu/~graves/preprint.pdf)
        - [A Theoretically Grounded Application of Dropout in Recurrent Neural
          Networks](http://arxiv.org/abs/1512.05287)

    Polyaxonfile usage:

    ```yaml
    LSTM:
      units: 3
      activation: tanh
    ```
    """
    IDENTIFIER = 'LSTM'
    SCHEMA = LSTMSchema

    def __init__(self,
                 units,
                 activation='tanh',
                 recurrent_activation='hard_sigmoid',
                 use_bias=True,
                 kernel_initializer=GlorotUniformInitializerConfig(),
                 recurrent_initializer=OrthogonalInitializerConfig,
                 bias_initializer=ZerosInitializerConfig(),
                 unit_forget_bias=True,
                 kernel_regularizer=None,
                 recurrent_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 recurrent_constraint=None,
                 bias_constraint=None,
                 dropout=0.,
                 recurrent_dropout=0.,
                 **kwargs):
        super(LSTMConfig, self).__init__(**kwargs)
        self.units = units
        self.activation = activation
        self.recurrent_activation = recurrent_activation
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.recurrent_initializer = recurrent_initializer
        self.bias_initializer = bias_initializer
        self.unit_forget_bias = unit_forget_bias
        self.kernel_regularizer = kernel_regularizer
        self.recurrent_regularizer = recurrent_regularizer
        self.bias_regularizer = bias_regularizer
        self.activity_regularizer = activity_regularizer
        self.kernel_constraint = kernel_constraint
        self.recurrent_constraint = recurrent_constraint
        self.bias_constraint = bias_constraint
        self.dropout = dropout
        self.recurrent_dropout = recurrent_dropout
